![](https://github.com/Narsil/spm_precompiled/workflows/build/badge.svg)
[![Crate](https://img.shields.io/crates/v/spm_precompiled.svg)](https://crates.io/crates/spm_precompiled)
[![API](https://docs.rs/spm_precompiled/badge.svg)](https://docs.rs/spm_precompiled)

# {{crate}}

{{readme}}
